import dataclasses
from typing import Any, List, Optional, Union, cast

import click
import fmf.utils

import tmt.base
import tmt.result
import tmt.steps
import tmt.steps.discover.fmf
import tmt.steps.execute
import tmt.steps.provision
import tmt.utils
from tmt.steps.discover import DiscoverPlugin
from tmt.steps.discover.fmf import DiscoverFmf, DiscoverFmfStepData
from tmt.steps.execute import ExecutePlugin
from tmt.steps.execute.internal import ExecuteInternal
from tmt.utils import Path

STATUS_VARIABLE = 'IN_PLACE_UPGRADE'
BEFORE_UPGRADE_PREFIX = 'old'
DURING_UPGRADE_PREFIX = 'upgrade'
AFTER_UPGRADE_PREFIX = 'new'
UPGRADE_DIRECTORY = 'upgrade'

PROPAGATE_TO_DISCOVER_KEYS = ['url', 'ref', 'filter', 'test', 'exclude']


@dataclasses.dataclass
class ExecuteUpgradeData(tmt.steps.execute.ExecuteStepData):
    url: Optional[str] = None
    upgrade_path: Optional[str] = None

    # Inherit from tmt.steps.discover.fmf.DiscoverFmfStepData
    ref: Optional[str] = None
    filter: Optional[List[str]] = None
    test: Optional[List[str]] = None
    exclude: Optional[List[str]] = None

    _normalize_test = tmt.utils.NormalizeKeysMixin._normalize_string_list
    _normalize_filter = tmt.utils.NormalizeKeysMixin._normalize_string_list
    _normalize_exclude = tmt.utils.NormalizeKeysMixin._normalize_string_list


@tmt.steps.provides_method('upgrade')
class ExecuteUpgrade(ExecuteInternal):
    """
    Perform system upgrade during testing.

    The upgrade executor runs the discovered tests (using the internal
    executor), then performs a set of upgrade tasks from a remote
    repository, and finally, re-runs the tests on the upgraded guest.

    The IN_PLACE_UPGRADE environment variable is set during the test
    execution to differentiate between the stages of the test. It is set
    to "old" during the first execution and "new" during the second
    execution. Test names are prefixed with this value to make the names
    unique.

    The upgrade tasks performing the actual system upgrade are taken
    from a remote repository either based on an upgrade path
    (e.g. fedora35to36) or filters. The upgrade path must correspond to
    a plan name in the remote repository whose discover step selects
    tests (upgrade tasks) performing the upgrade. Currently, selection
    of upgrade tasks in the remote repository can be done using both fmf
    and shell discover method. The supported keys in discover are:

      - ref
      - filter
      - exclude
      - tests
      - test

    The environment variables defined in the remote upgrade path plan are
    passed to the upgrade tasks when they are executed. An example of an
    upgrade path plan (in the remote repository):

        discover: # Selects appropriate upgrade tasks (L1 tests)
            how: fmf
            filter: "tag:fedora"
        environment: # This is passed to upgrade tasks
            SOURCE: 35
            TARGET: 36
        execute:
            how: tmt

    If no upgrade path is specified in the plan, the tests (upgrade tasks)
    are selected based on the configuration of the upgrade plugin
    (e.g. based on the filter in its configuration).

    If these two possible ways of specifying upgrade tasks are combined,
    the remote discover plan is used but its options are overridden
    with the values specified locally.

    The same options and config keys and values can be used as in the
    internal executor.

    Minimal execute config example with an upgrade path:

        execute:
            how: upgrade
            url: https://github.com/teemtee/upgrade
            upgrade-path: /paths/fedora35to36

    Execute config example without an upgrade path:

        execute:
            how: upgrade
            url: https://github.com/teemtee/upgrade
            filter: "tag:fedora"
    """

    # FIXME: ignore[assignment]: https://github.com/teemtee/tmt/issues/1540
    _data_class = ExecuteUpgradeData  # type: ignore[assignment]

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._discover_upgrade: Optional[DiscoverFmf] = None

    @classmethod
    def options(cls, how: Optional[str] = None) -> List[tmt.options.ClickOptionDecoratorType]:
        """ Prepare command line options for given method """
        return [
            click.option(
                '--url', '-u', metavar='REPOSITORY',
                help='URL of the git repository with upgrade tasks.'),
            click.option(
                '--upgrade-path', '-p', metavar='PLAN_NAME',
                help='Upgrade path corresponding to a plan name in the repository '
                'with upgrade tasks.'),
            tmt.steps.discover.fmf.REF_OPTION,
            tmt.steps.discover.fmf.TEST_OPTION,
            tmt.steps.discover.fmf.FILTER_OPTION,
            tmt.steps.discover.fmf.EXCLUDE_OPTION
            ] + super().options(how)

    @property
    def discover(self) -> Union[tmt.steps.discover.Discover, DiscoverFmf]:  # type:ignore[override]
        """ Return discover plugin instance """
        # If we are in the second phase (upgrade), take tests from our fake
        # discover plugin.
        if self._discover_upgrade:
            return self._discover_upgrade
        else:
            return self.step.plan.discover

    def go(
            self,
            *,
            guest: 'tmt.steps.provision.Guest',
            environment: Optional[tmt.utils.EnvironmentType] = None,
            logger: tmt.log.Logger) -> None:
        """ Execute available tests """
        self._results: List[tmt.result.Result] = []
        # Inform about the how, skip the actual execution
        ExecutePlugin.go(self, guest=guest, environment=environment, logger=logger)

        self.url = self.get('url')
        self.upgrade_path = self.get('upgrade-path')
        if not self.url:
            raise tmt.utils.ExecuteError(
                "URL to repository with upgrade tasks must be specified.")
        for key in self._keys:
            value = self.get(key)
            if value:
                if key == "test":
                    self.info('test', fmf.utils.listed(value), 'green')
                else:
                    self.info(key, value, color='green')

        # Nothing to do in dry mode
        if self.opt('dry'):
            self._results = []
            return

        self.verbose(
            'upgrade', 'run tests on the old system', color='blue', shift=1)
        self._run_test_phase(guest, BEFORE_UPGRADE_PREFIX)
        self.verbose(
            'upgrade', 'perform the system upgrade', color='blue', shift=1)
        self._perform_upgrade(guest)
        self.verbose(
            'upgrade', 'run tests on the new system', color='blue', shift=1)
        self._run_test_phase(guest, AFTER_UPGRADE_PREFIX)

    def _get_plan(self, upgrades_repo: Path) -> tmt.base.Plan:
        """ Get plan based on upgrade path """
        tree = tmt.base.Tree(logger=self._logger, path=upgrades_repo)
        try:
            # We do not want to consider plan -n provided on the command line
            # in the remote repo for finding upgrade path.
            tmt.base.Plan.ignore_class_options = True
            plans = tree.plans(names=[self.upgrade_path])
        finally:
            tmt.base.Plan.ignore_class_options = False

        if len(plans) == 0:
            raise tmt.utils.ExecuteError(
                f"No matching upgrade path found for '{self.upgrade_path}'.")
        elif len(plans) > 1:
            names = [plan.name for plan in plans]
            raise tmt.utils.ExecuteError(
                f"Ambiguous upgrade path reference, found plans "
                f"{fmf.utils.listed(names)}.")
        return plans[0]

    def _fetch_upgrade_tasks(self) -> None:
        """ Fetch upgrade tasks using DiscoverFmf """
        data = DiscoverFmfStepData(
            name='upgrade-discover',
            how='fmf',
            # url=self.data.url,
            **{
                key: getattr(self.data, key) for key in PROPAGATE_TO_DISCOVER_KEYS
                }
            )

        self._discover_upgrade = DiscoverFmf(logger=self._logger, step=self.step, data=data)
        self._run_discover_upgrade()

    def _run_discover_upgrade(self) -> None:
        """ Silently run discover upgrade """
        # Make it quiet, we do not want any output from discover
        assert self._discover_upgrade is not None
        assert self._discover_upgrade._context is not None
        quiet = self._discover_upgrade._context.params['quiet']
        try:
            self._discover_upgrade._context.params['quiet'] = True
            # Discover normally uses also options from global Test class
            # (e.g. test -n foo). Ignore this when selecting upgrade tasks.
            tmt.base.Test.ignore_class_options = True
            self._discover_upgrade.wake()
            self._discover_upgrade.go()
        finally:
            self._discover_upgrade._context.params['quiet'] = quiet
            tmt.base.Test.ignore_class_options = False

    def _prepare_remote_discover_data(self, plan: tmt.base.Plan) -> tmt.steps._RawStepData:
        """ Merge remote discover data with the local filters """
        if len(plan.discover.data) > 1:
            raise tmt.utils.ExecuteError(
                "Multiple discover configs are not supported.")

        data = plan.discover.data[0]

        remote_raw_data: tmt.steps._RawStepData = {
            # Force name
            'name': 'upgrade-discover-remote',
            'how': 'fmf'
            }
        remote_raw_data.update(cast(tmt.steps._RawStepData, {
            tmt.utils.key_to_option(key): value
            for key, value in data.items()
            if key in PROPAGATE_TO_DISCOVER_KEYS
            }))

        # Local values have priority, override
        for key in self._keys:
            value = self.get(key)
            if key in PROPAGATE_TO_DISCOVER_KEYS and value:
                remote_raw_data[key] = value  # type:ignore[literal-required]

        return remote_raw_data

    def _perform_upgrade(self, guest: tmt.steps.provision.Guest) -> None:
        """ Perform a system upgrade """
        try:
            self._fetch_upgrade_tasks()
            extra_environment = None
            assert self._discover_upgrade is not None
            if self.upgrade_path:
                # Create a fake discover from the data in the upgrade path
                plan = self._get_plan(self._discover_upgrade.testdir)
                data = self._prepare_remote_discover_data(plan)
                # Unset `url` because we don't want discover step to perform clone.
                # Instead, we want it to re-use existing, already cloned path.
                # ignore[typeddict-item]: data is _RwStepData, we do not have more detailed type
                # for raw step data of internal/upgrade plugins, it would be pretty verbose.
                data['url'] = None  # type: ignore[typeddict-item]
                data['path'] = self._discover_upgrade.testdir  # type: ignore[typeddict-item]
                # FIXME: cast() - https://github.com/teemtee/tmt/issues/1599
                self._discover_upgrade = cast(DiscoverFmf, DiscoverPlugin.delegate(
                    self.step, raw_data=data))
                self._run_discover_upgrade()
                # Pass in the path-specific env variables
                extra_environment = plan.environment
            for test in self._discover_upgrade.tests():
                test.name = f'/{DURING_UPGRADE_PREFIX}/{test.name.lstrip("/")}'
            self._run_tests(guest, extra_environment=extra_environment)
        finally:
            self._discover_upgrade = None

    def _run_test_phase(self, guest: tmt.steps.provision.Guest, prefix: str) -> None:
        """
        Execute a single test phase on the guest

        Tests names are prefixed with the prefix argument in order to make
        their names unique so that the results are distinguishable.
        The prefix is also set as IN_PLACE_UPGRADE environment variable.
        """
        names_backup = []
        for test in self.discover.tests():
            names_backup.append(test.name)
            test.name = f'/{prefix}/{test.name.lstrip("/")}'

        self._run_tests(guest, extra_environment={STATUS_VARIABLE: prefix})

        tests = self.discover.tests()
        for i, test in enumerate(tests):
            test.name = names_backup[i]
