import dataclasses
from typing import Any, Dict, List, Optional, cast

import fmf

import tmt
import tmt.steps
import tmt.steps.prepare
import tmt.utils
from tmt.steps.provision import Guest
from tmt.utils import ShellScript, field


# TODO: remove `ignore` with follow-imports enablement
@dataclasses.dataclass
class PrepareShellData(tmt.steps.prepare.PrepareStepData):
    script: List[ShellScript] = field(
        default_factory=list,
        option=('-s', '--script'),
        multiple=True,
        metavar='SCRIPT',
        help='Shell script to be executed. Can be used multiple times.',
        normalize=tmt.utils.normalize_shell_script_list
        )

    # ignore[override] & cast: two base classes define to_spec(), with conflicting
    # formal types.
    def to_spec(self) -> Dict[str, Any]:  # type: ignore[override]
        data = cast(Dict[str, Any], super().to_spec())
        data['script'] = [str(script) for script in self.script]

        return data

    def to_serialized(self) -> Dict[str, Any]:
        data = super().to_serialized()

        data['script'] = [str(script) for script in self.script]

        return data

    @classmethod
    def from_serialized(cls, serialized: Dict[str, Any]) -> 'PrepareShellData':
        """ Convert from a serialized form loaded from a file """

        obj = super().from_serialized(serialized)
        obj.script = [ShellScript(script) for script in serialized['script']]

        return obj


# TODO: drop ignore once type annotations between modules enabled
@tmt.steps.provides_method('shell')
class PrepareShell(tmt.steps.prepare.PreparePlugin):
    """
    Prepare guest using shell (bash) scripts

    Example config:

    prepare:
        how: shell
        script:
        - sudo dnf install -y 'dnf-command(copr)'
        - sudo dnf copr enable -y psss/tmt
        - sudo dnf install -y tmt

    Use 'order' attribute to select in which order preparation should
    happen if there are multiple configs. Default order is '50'.
    Default order of required packages installation is '70'.
    """

    _data_class = PrepareShellData

    def go(
            self,
            *,
            guest: 'Guest',
            environment: Optional[tmt.utils.EnvironmentType] = None,
            logger: tmt.log.Logger) -> None:
        """ Prepare the guests """
        super().go(guest=guest, environment=environment, logger=logger)

        # Give a short summary
        scripts: List[tmt.utils.ShellScript] = self.get('script')
        overview = fmf.utils.listed(scripts, 'script')
        logger.info('overview', f'{overview} found', 'green')

        # Execute each script on the guest (with default shell options)
        for script in scripts:
            logger.verbose('script', str(script), 'green')
            script_with_options = tmt.utils.ShellScript(f'{tmt.utils.SHELL_OPTIONS}; {script}')
            guest.execute(script_with_options, cwd=self.step.plan.worktree)
